//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 04: OpenMp
//
//  Description: Collatz in openMP
//  	omp06Collatz.c: the hello work openmp version
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc omp06Collatz.c -o omp06Collatz -fopenmp
Executing:	./omp06Collatz

Output in my Laptop:

Steps 174, Frequency 1031401

real	0m6.241s
user	0m49.663s
sys	0m0.048s

*/
//************************************************************


// *******************************************************************
#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline")
#pragma GCC option("arch=native","tune=native","no-zero-upper")
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <omp.h>


// Maximum natural number
//#define N         10000000
#define N          100000000
//#define N       1000000000


// Maximum number of steps
#define NSTEPS  100000
unsigned long Hist[NSTEPS+1] ;

#define DEPTH	1000000
unsigned long Pass[DEPTH] ;


// ***************************************
unsigned long CollatzOne( unsigned long st )
{
	unsigned long k = 0 ;
	unsigned long q = st ;
        for ( k = 0 ; k < NSTEPS ; k++ ) {
		if ( q < DEPTH )
			if ( Pass[q] )
				return k + Pass[q] ;
        	if ( q == 1 )
        		break ;
        	if ( q%2 == 0 )
        		q /= 2 ;
        	else
        		q=3*q+1;
	}
	if ( st < DEPTH )
		Pass[st] = k ;
	return k ;
}


// ***************************************
void CollatzFast(void)
{

#pragma omp parallel for schedule(dynamic,1000)
	for( unsigned long i = 1 ; i < N ; i++  ) {
	unsigned long k = 0 ;
	unsigned long q = i ;
        for ( k = 0 ; k < NSTEPS ; k++ ) {
		if ( q < DEPTH )
			if ( Pass[q] ) {
				k += Pass[q] ;
				break ;
			}
        	if ( q == 1 )
        		break ;
        	if ( q%2 == 0 )
        		q /= 2 ;
        	else
        		q=3*q+1;
	}
	if ( i < DEPTH )
//#pragma omp critical
		Pass[i] = k ;

	if ( k < NSTEPS )
#pragma omp atomic
               	Hist[k]++ ;
	else
#pragma omp atomic
		Hist[NSTEPS]++ ;
    }
}

// ***************************************
void PrintResults()
{
    unsigned long int max=0 ;
    unsigned long ind=0;

for ( int k = 0 ; k < NSTEPS ; k++ )
    if ( max < Hist[k] ) {
        ind = k ;
        max = Hist[k] ;
    }
    printf( "Steps %lu, Frequency %lu\n", ind, max) ;
}

// ***************************************
void PrintPass()
{

for ( int k = 0 ; k < DEPTH ; k++ )
    printf( "Nm %d, Steps %lu\n", k, Pass[k]) ;
}


// ***************************************
int main()
{
//    omp_set_num_threads(8);
    CollatzFast() ;
    PrintResults() ;
//	PrintPass() ;
    printf( "Nm %d, Steps %lu\n", NSTEPS, Pass[NSTEPS]) ;
    return 0 ;
}
